/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __math_DAGNode_h__
#define __math_DAGNode_h__

namespace fe
{

#if FE_CODEGEN<=FE_DEBUG
#define FE_DAG_VALIDATE()	//validate()
#else
#define FE_DAG_VALIDATE()
#endif

/**************************************************************************//**
	@brief Node in a Directed Acyclic Graph

	@ingroup math
*//***************************************************************************/
class FE_DL_EXPORT DAGNode: public Handled<DAGNode>
{
	public:

	class Connection: public Handled<Connection>
	{
		public:
						Connection(sp<DAGNode> a_spChild,
								hp<DAGNode> a_hpParent,
								const String a_childConnector,
								const String a_parentConnector):
							m_spChild(a_spChild),
							m_hpParent(a_hpParent),
							m_childConnector(a_childConnector),
							m_parentConnector(a_parentConnector)			{}
			bool		operator==(const sp<Connection>& a_spOther)
						{	return a_spOther->m_spChild.raw()==
									m_spChild.raw(); }
			sp<DAGNode>	child(void)				{ return m_spChild; }
			hp<DAGNode>	parent(void)			{ return m_hpParent; }
			String&		childConnector(void)	{ return m_childConnector; }
			String&		parentConnector(void)	{ return m_parentConnector; }

		private:
			sp<DAGNode>	m_spChild;
			hp<DAGNode>	m_hpParent;
			String		m_childConnector;
			String		m_parentConnector;
	};

						DAGNode(void)
						{
#if FE_COUNTED_TRACK
							setName("DAGNode");
							registerRegion(this,sizeof(*this));
#endif
						}

virtual					~DAGNode(void)
						{
#if FE_COUNTED_TRACK
							deregisterRegion(this);
#endif
						}

		void			addParentConnector(const String a_connector);
		void			addChildConnector(const String a_connector);

		List<String>::Iterator parentConnector(void)
						{	return m_parentConnectors; }
		List<String>::Iterator childConnector(void)
						{	return m_childConnectors; }

		U32				parentConnectorCount(void) const
						{	return m_parentConnectors.size(); }
		U32				childConnectorCount(void) const
						{	return m_childConnectors.size(); }

						//* NOTE index operations on a list can be slow
		String			parentConnector(U32 a_index)
						{	return m_parentConnectors[a_index]; }
		String			childConnector(U32 a_index)
						{	return m_childConnectors[a_index]; }

		BWORD			hasParentConnector(String a_connector);
		BWORD			hasChildConnector(String a_connector);

		BWORD			attachTo(sp<DAGNode> rNode,String a_localConnector="",
								String a_remoteConnector="");
		BWORD			detachFrom(sp<DAGNode> rNode);
		void			detach(void);
		BWORD			detach(String a_localConnector);
		void			validate(void);

		List< hp<Connection> >::Iterator parentConnection(void)
						{	return m_parentConnections; }
		List< sp<Connection> >::Iterator childConnection(void)
						{	return m_childConnections; }

		U32				parentConnectionCount(void) const
						{	return m_parentConnections.size(); }
		U32				childConnectionCount(void) const
						{	return m_childConnections.size(); }

						//* NOTE index operations on a list can be slow
		sp<Connection>	parentConnection(U32 a_index) const
						{	return m_parentConnections[a_index]; }
		sp<Connection>	childConnection(U32 a_index) const
						{	return m_childConnections[a_index]; }

		sp<Connection>	parentConnection(String a_parentConnector) const;
		sp<Connection>	childConnection(String a_childConnector) const;

	private:

		List<String>			m_parentConnectors;
		List<String>			m_childConnectors;

		List< hp<Connection> >	m_parentConnections;
		List< sp<Connection> >	m_childConnections;

		std::map< String, hp<Connection> >	m_parentMap;
		std::map< String, sp<Connection> >	m_childMap;
};

inline void DAGNode::addParentConnector(const String a_connector)
{
	ListCore::Context context;
	if(!m_parentConnectors.searchForElement(context,a_connector))
	{
		m_parentConnectors.append(a_connector);
	}
}

inline void DAGNode::addChildConnector(const String a_connector)
{
	ListCore::Context context;
	if(!m_childConnectors.searchForElement(context,a_connector))
	{
		m_childConnectors.append(a_connector);
	}
}

inline BWORD DAGNode::hasParentConnector(String a_connector)
{
	ListCore::Context context;
	return m_parentConnectors.searchForElement(context,a_connector);
}

inline BWORD DAGNode::hasChildConnector(String a_connector)
{
	ListCore::Context context;
	return m_childConnectors.searchForElement(context,a_connector);
}

} /* namespace */

#endif /* __math_DAGNode_h__ */
