/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

#define FE_SAL_DEBUG		FALSE
#define FE_SAL_PICK_DEBUG	FALSE

namespace fe
{

void SurfaceAttrLabOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* page Config
	catalog<String>("Name")="uv";
	catalog<String>("Name","page")="Config";

	//* page Edit
	catalog<bool>("Saturate")=false;
	catalog<String>("Saturate","page")="Edit";
	catalog<bool>("Saturate","joined")=true;

	catalog<Real>("SaturateCenter")=0.5;
	catalog<String>("SaturateCenter","label")="Threshold";
	catalog<String>("SaturateCenter","enabler")="Saturate";
	catalog<String>("SaturateCenter","page")="Edit";

	catalog<Real>("SaturateToMin")=0.0;
	catalog<Real>("SaturateToMin","high")=1.0;
	catalog<String>("SaturateToMin","label")="To Min";
	catalog<String>("SaturateToMin","enabler")="Saturate";
	catalog<String>("SaturateToMin","page")="Edit";
	catalog<bool>("SaturateToMin","joined")=true;

	catalog<Real>("SaturateToMax")=1.0;
	catalog<Real>("SaturateToMax","high")=1.0;
	catalog<String>("SaturateToMax","label")="To Max";
	catalog<String>("SaturateToMax","enabler")="Saturate";
	catalog<String>("SaturateToMax","page")="Edit";

	catalog<bool>("Gradient")=false;
	catalog<String>("Gradient","page")="Edit";
	catalog<bool>("Gradient","joined")=true;

	catalog<Real>("GradientMax")=1.0;
	catalog<Real>("GradientMax","high")=1.0;
	catalog<String>("GradientMax","label")="Threshold";
	catalog<String>("GradientMax","page")="Edit";
	catalog<String>("GradientMax","enabler")="Gradient";

	catalog<I32>("GradientIterations")=1;
	catalog<String>("GradientIterations","label")="Iterations";
	catalog<String>("GradientIterations","page")="Edit";
	catalog<String>("GradientIterations","enabler")="Gradient";

	catalog<bool>("Smooth")=false;
	catalog<String>("Smooth","page")="Edit";
	catalog<bool>("Smooth","joined")=true;

	catalog<Real>("Smoothing")=1.0;
	catalog<Real>("Smoothing","high")=1.0;
	catalog<String>("Smoothing","page")="Edit";
	catalog<String>("Smoothing","enabler")="Smooth";

	catalog<I32>("SmoothIterations")=1;
	catalog<String>("SmoothIterations","label")="Iterations";
	catalog<String>("SmoothIterations","page")="Edit";
	catalog<String>("SmoothIterations","enabler")="Smooth";

	//* page Display
	catalog<bool>("DisplayValue")=true;
	catalog<String>("DisplayValue","label")="Value Text";
	catalog<String>("DisplayValue","page")="Display";

	catalog<bool>("DisplayMag")=true;
	catalog<String>("DisplayMag","label")="Mag Guides";
	catalog<String>("DisplayMag","page")="Display";
	catalog<bool>("DisplayMag","joined")=true;

	catalog<Real>("DisplayMagScale")=1.0;
	catalog<String>("DisplayMagScale","label")="Scale";
	catalog<String>("DisplayMagScale","enabler")="DisplayMag";
	catalog<String>("DisplayMagScale","page")="Display";

	catalog<bool>("DisplayShape")=true;
	catalog<String>("DisplayShape","label")="Shape";
	catalog<String>("DisplayShape","page")="Display";
	catalog<bool>("DisplayShape","joined")=true;

	catalog<Real>("DisplayShapeRadius")=0.1;
	catalog<String>("DisplayShapeRadius","label")="Radius";
	catalog<String>("DisplayShapeRadius","enabler")="DisplayShape";
	catalog<String>("DisplayShapeRadius","page")="Display";
	catalog<bool>("DisplayShapeRadius","joined")=true;

	catalog<bool>("DisplayShapeCentered")=false;
	catalog<String>("DisplayShapeCentered","label")="Centered";
	catalog<String>("DisplayShapeCentered","enabler")="DisplayShape";
	catalog<String>("DisplayShapeCentered","page")="Display";
	catalog<bool>("DisplayShapeCentered","joined")=true;

	catalog<Real>("DisplayShapeLength")=1.0;
	catalog<String>("DisplayShapeLength","label")="Length";
	catalog<String>("DisplayShapeLength","enabler")="DisplayShape";
	catalog<String>("DisplayShapeLength","page")="Display";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","visible")="any";

	catalog< sp<Component> >("Input Surface");

	m_spSolid=new DrawMode();
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setBackfaceCulling(FALSE);
	m_spSolid->setLineWidth(1.0);
}

void SurfaceAttrLabOp::handle(Record& a_rSignal)
{
#if FE_SAL_DEBUG
	feLog("SurfaceAttrLabOp::handle\n");
#endif

	const String attrName=catalog<String>("Name");
	if(attrName.empty())
	{
		feLog("SurfaceAttrLabOp::handle no name\n");
		return;
	}

	SpatialVector line[4];

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

	if(spDrawBrush.isValid())
	{
		spDrawBrush->pushDrawMode(m_spSolid);
		spDrawOverlay->pushDrawMode(m_spSolid);

		if(m_spInput.isValid())
		{
			const Color black(0.0,0.0,0.0,1.0);
			const Color red(0.6,0.0,0.0,1.0);
			const Color green(0.0,0.6,0.0,1.0);
			const Color blue(0.0,0.0,0.6,1.0);
			const Color cyan(0.4,1.0,1.0,1.0);
			const Color magenta(1.0,0.4,1.0,1.0);
			const Color yellow(1.0,1.0,0.0,1.0);
			const Color white(1.0,1.0,1.0,1.0);

			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			const String attrType=m_spOutputElement->type();

#if FE_SAL_PICK_DEBUG
			feLog("SurfaceAttrLabOp::handle ray %s  %s\n",
					c_print(rRayOrigin),c_print(rRayDirection));
#endif
			const Real maxDistance=100.0;
			sp<SurfaceI::ImpactI> spImpact=m_spInput->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);
			if(spImpact.isValid())
			{
#if FE_SAL_PICK_DEBUG
				feLog("SurfaceAttrLabOp::handle hit\n");
#endif
				const BWORD displayValue=catalog<bool>("DisplayValue");

				m_event.bind(windowEvent(a_rSignal));
//				feLog("SurfaceAttrLabOp::handle %s\n",c_print(m_event));

				sp<ViewI> spView=spDrawOverlay->view();

				I32 pickIndex= -1;

				sp<SurfaceCurves::Impact> spCurveImpact=spImpact;
				if(spCurveImpact.isValid())
				{
					pickIndex=spCurveImpact->pickIndex();

#if FE_SAL_PICK_DEBUG
					feLog("  curve pick %d\n",pickIndex);
#endif
				}

				sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
				if(spTriImpact.isValid())
				{
					pickIndex=spTriImpact->pickIndex();

#if FE_SAL_PICK_DEBUG
					feLog("  triangle pick %d\n",pickIndex);
#endif
				}

				if(pickIndex>=0)
				{
					SpatialVector average;
					set(average);

					const U32 subCount=m_spOutputVertices->subCount(pickIndex);
					for(I32 subIndex=0;subIndex<subCount;subIndex++)
					{
						const I32 pointIndex=
								m_spOutputVertices->integer(pickIndex,subIndex);
						const SpatialVector point=
								m_spOutputPoint->spatialVector(pointIndex);

						average+=point;

						const Vector2i projected=
								spView->project(point,ViewMode::e_perspective);

						if(displayValue)
						{
							String text;

							if(attrType=="String")
							{
								text=m_spOutputElement->string(pointIndex);
							}
							else if(attrType=="Real")
							{
								const Real realValue=
										m_spOutputElement->real(pointIndex);

								text.sPrintf("%.3G",realValue);
							}
							else if(attrType=="Integer")
							{
								const I32 intValue=
										m_spOutputElement->integer(pointIndex);

								text.sPrintf("%d",intValue);
							}
							else
							{
								const SpatialVector vectorValue=
										m_spOutputElement->spatialVector(
										pointIndex);

								text.sPrintf("%s",c_print(vectorValue));
							}

							//* TODO why is this bg showing up black?
							drawLabel(spDrawOverlay,projected,
									text,0,white,black,&blue);

							String text2;
							text2.sPrintf("%d",pointIndex);
							drawLabel(spDrawOverlay,projected+Vector2i(0,9),
									text2,0,cyan,black,&black);
						}
					}
					if(subCount && displayValue)
					{
						average*=1.0/subCount;

						String text;
						text.sPrintf("%d",pickIndex);

						const Vector2i projected=spView->project(
								average,ViewMode::e_perspective);

						drawLabel(spDrawOverlay,projected,
								text,0,magenta,black,&black);
					}
				}
			}

			if(catalog<bool>("DisplayShape"))
			{
				const Real displayShapeRadius=
						catalog<Real>("DisplayShapeRadius");
				const Real displayShapeLength=
						catalog<Real>("DisplayShapeLength");
				const BWORD displayShapeCentered=
						catalog<bool>("DisplayShapeCentered");

				if(attrType=="Real")
				{
					const I32 primitiveCount=m_spOutputVertices->count();
					for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;
							primitiveIndex++)
					{
						const U32 subCount=
								m_spOutputVertices->subCount(primitiveIndex);
						for(I32 subIndex=0;subIndex<subCount;subIndex++)
						{
							const I32 pointIndex=m_spOutputVertices->integer(
									primitiveIndex,subIndex);
							const SpatialVector point=
									m_spOutputPoint->spatialVector(pointIndex);
							const Real value=m_spOutputElement->real(pointIndex);

							SpatialTransform sphereTransform;
							setIdentity(sphereTransform);
							setTranslation(sphereTransform,point);

							const Real sphereRadius=displayShapeRadius*value;
							const SpatialVector sphereScale(sphereRadius,
									sphereRadius,sphereRadius);

							spDrawBrush->drawSphere(sphereTransform,
									&sphereScale,yellow);
						}
					}
				}
				else if(attrType.match("Vector."))
				{
					const I32 primitiveCount=m_spOutputVertices->count();
					for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;
							primitiveIndex++)
					{
						const U32 subCount=
								m_spOutputVertices->subCount(primitiveIndex);
						for(I32 subIndex=0;subIndex<subCount;subIndex++)
						{
							const I32 pointIndex=m_spOutputVertices->integer(
									primitiveIndex,subIndex);
							const SpatialVector point=
									m_spOutputPoint->spatialVector(pointIndex);
							const SpatialVector value=displayShapeLength*
									m_spOutputElement->spatialVector(
									pointIndex);

							SpatialVector location1;
							SpatialVector location2;
							if(displayShapeCentered)
							{
								location1=point-0.5*value;
								location2=point+0.5*value;
							}
							else
							{
								location1=point;
								location2=point+value;
							}

							const Real radius1=displayShapeRadius;
							const Real radius2=displayShapeRadius*0.01;
							const U32 slices=8;

							spDrawBrush->drawCylinder(location1,location2,
								radius1,radius2,yellow,slices);
						}
					}
				}
			}
		}

		m_keepInput=TRUE;

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	if(!m_keepInput)
	{
		if(!access(m_spInput,"Input Surface")) return;
	}
	m_keepInput=FALSE;

	if(!accessOutput(m_spOutputPoint,a_rSignal,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputNormal;
	if(!accessOutput(spOutputNormal,a_rSignal,e_point,e_normal)) return;

	if(!accessOutput(m_spOutputElement,a_rSignal,e_point,attrName)) return;

	if(!accessOutput(m_spOutputVertices,a_rSignal,
			e_primitive,e_vertices)) return;

	const I32 elementCount=m_spOutputElement->count();
	const I32 normalCount=spOutputNormal->count();
	const String attrType=m_spOutputElement->type();
	const BWORD isVector=attrType.match("Vector.");

#if FE_SAL_DEBUG
	feLog("SurfaceAttrLabOp::handle type %s elementCount %d normalCount %d\n",
			attrType,elementCount,normalCount);
#endif

	if(catalog<bool>("Saturate"))
	{
		const Real saturateCenter=catalog<Real>("SaturateCenter");
		const Real saturateToMin=catalog<Real>("SaturateToMin");
		const Real saturateToMax=catalog<Real>("SaturateToMax");

		for(I32 index=0;index<elementCount;index++)
		{
			if(isVector)
			{
				SpatialVector value=m_spOutputElement->spatialVector(index);
				const Real mag=magnitude(value);
				if(mag>0.0)
				{
					const Real newMag=
							mag<saturateCenter? saturateToMin: saturateToMax;
					value*=newMag/mag;
					m_spOutputElement->set(index,value);
				}
			}
			else
			{
				const Real mag=m_spOutputElement->real(index);

				m_spOutputElement->set(index,
						mag<saturateCenter? saturateToMin: saturateToMax);
			}
		}
	}

	if(catalog<bool>("Gradient"))
	{
		const Real gradientMax=catalog<Real>("GradientMax");
		const Real gradientMax2=gradientMax*gradientMax;
		const I32 gradientIterations=catalog<I32>("GradientIterations");

		const I32 primitiveCount=m_spOutputVertices->count();

		for(I32 pass=0;pass<gradientIterations;pass++)
		{
			for(I32 index=0;index<primitiveCount;index++)
			{
				const U32 subCount=m_spOutputVertices->subCount(index);
				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const U32 lastSubIndex=subIndex? subIndex-1: subCount-1;

					const I32 vertexIndexA=
							m_spOutputVertices->integer(index,lastSubIndex);
					const I32 vertexIndexB=
							m_spOutputVertices->integer(index,subIndex);

					if(isVector)
					{
						SpatialVector elementA=
								m_spOutputElement->spatialVector(vertexIndexA);
						SpatialVector elementB=
								m_spOutputElement->spatialVector(vertexIndexB);

						const SpatialVector delta=elementB-elementA;
						const Real dist2=magnitudeSquared(delta);
						if(dist2<=gradientMax2)
						{
							continue;
						}
						const Real dist=sqrt(dist2);

						const SpatialVector adjust=
							0.5*(dist-gradientMax)/dist*delta;

						m_spOutputElement->set(vertexIndexA,elementA+adjust);
						m_spOutputElement->set(vertexIndexB,elementB-adjust);
					}
					else
					{
						Real elementA=m_spOutputElement->real(vertexIndexA);
						Real elementB=m_spOutputElement->real(vertexIndexB);

						const Real rise=elementB-elementA;
						if(fabs(rise)<=gradientMax)
						{
							continue;
						}

						const Real adjust=0.5*
								(rise>0? rise-gradientMax: rise+gradientMax);

						m_spOutputElement->set(vertexIndexA,elementA+adjust);
						m_spOutputElement->set(vertexIndexB,elementB-adjust);
					}
				}
			}
		}
	}

	if(catalog<bool>("Smooth"))
	{
		const Real smoothing=catalog<Real>("Smoothing");
		const I32 smoothIterations=catalog<I32>("SmoothIterations");

		const I32 primitiveCount=m_spOutputVertices->count();

		for(I32 pass=0;pass<smoothIterations;pass++)
		{
			for(I32 index=0;index<primitiveCount;index++)
			{
				const U32 subCount=m_spOutputVertices->subCount(index);
				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const U32 lastSubIndex=subIndex? subIndex-1: subCount-1;

					const I32 vertexIndexA=
							m_spOutputVertices->integer(index,lastSubIndex);
					const I32 vertexIndexB=
							m_spOutputVertices->integer(index,subIndex);

					if(isVector)
					{
						SpatialVector elementA=
								m_spOutputElement->spatialVector(vertexIndexA);
						SpatialVector elementB=
								m_spOutputElement->spatialVector(vertexIndexB);

						const SpatialVector average=0.5*(elementA+elementB);

						elementA+=smoothing*(average-elementA);
						elementB+=smoothing*(average-elementB);

						m_spOutputElement->set(vertexIndexA,elementA);
						m_spOutputElement->set(vertexIndexB,elementB);
					}
					else
					{
						Real elementA=m_spOutputElement->real(vertexIndexA);
						Real elementB=m_spOutputElement->real(vertexIndexB);

						const Real average=0.5*(elementA+elementB);

						elementA+=smoothing*(average-elementA);
						elementB+=smoothing*(average-elementB);

						m_spOutputElement->set(vertexIndexA,elementA);
						m_spOutputElement->set(vertexIndexB,elementB);
					}
				}
			}
		}
	}

	if(catalog<bool>("DisplayMag"))
	{
		const Real displayMagScale=catalog<Real>("DisplayMagScale");

		sp<DrawI> spDrawGuide;
		if(!accessGuide(spDrawGuide,a_rSignal)) return;

		for(I32 index=0;index<elementCount;index++)
		{
			const Real mag=isVector?
					magnitude(m_spOutputElement->spatialVector(index)):
					m_spOutputElement->real(index);

			const SpatialVector norm=index<normalCount?
					spOutputNormal->spatialVector(index,0):
					SpatialVector(0.0,1.0,0.0);

#if FE_SAL_DEBUG
			feLog("SurfaceAttrLabOp::handle %d/%d mag %.6G norm %s\n",
					index,elementCount,mag,c_print(norm));
#endif

			SpatialVector line[2];
			line[0]=m_spOutputPoint->spatialVector(index,0);
			line[1]=line[0]+displayMagScale*mag*norm;

#if FE_SAL_DEBUG
			feLog("  line %s  %s\n",c_print(line[0]),c_print(line[1]));
#endif

			spDrawGuide->drawLines(line,NULL,2,DrawI::e_strip,false,NULL);
		}
	}

#if FE_SAL_DEBUG
	feLog("SurfaceAttrLabOp::handle done\n");
#endif
}

} // namespace
