/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceAccessibleI_h__
#define __surface_SurfaceAccessibleI_h__

namespace fe
{

class SurfaceAccessorI;

/**************************************************************************//**
	@brief Surface Element Access and Alteration

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleI: virtual public Component
{
	public:
		enum	Element
				{
					e_point,
					e_pointGroup,
					e_vertex,
					e_primitive,
					e_primitiveGroup,
					e_detail
				};
		enum	Attribute
				{
					e_generic,
					e_position,
					e_normal,
					e_vertices
					// TODO common attributes
				};
		enum	Message
				{
					e_quiet,
					e_warning,
					e_error
				};
		enum	Creation
				{
					e_refuseMissing,
					e_createMissing
				};
		enum	Threading
				{
					e_singleThread,
					e_multiThread
				};

	//* TODO not pure -> move out
	class ThreadingState: public Counted
	{
		public:
											ThreadingState(void):
												m_threading(e_singleThread)	{}
			void							set(SurfaceAccessibleI::Threading
													a_threading)
											{	m_threading=a_threading; }
			SurfaceAccessibleI::Threading	threading(void)
											{	return m_threading; }
		private:
			SurfaceAccessibleI::Threading	m_threading;
	};

								/// @brief Inform threading intentions
virtual	void					bind(sp<SurfaceAccessibleI::ThreadingState>
										a_spThreadingState)					=0;
								/// @brief Start a thread-unsafe action
virtual	void					lock(void)									=0;
								/// @brief Finish a thread-unsafe action
virtual	void					unlock(void)								=0;

								/// @brief Return true if surface is available
virtual	BWORD					isBound(void)								=0;

								/// @brief Return access to an attribute
virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										String a_name,
										SurfaceAccessibleI::Creation a_create
										)									=0;
virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute,
										SurfaceAccessibleI::Creation a_create
										)									=0;
virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										String a_name)						=0;
virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute)						=0;

								/// @brief Return current underlying SurfaceI
virtual	sp<SurfaceI>			surface(void)								=0;
virtual	sp<SurfaceI>			surface(String a_group)						=0;
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex)					=0;
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group)	=0;

virtual	void					append(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const SpatialTransform&
										a_rTransform)						=0;

virtual	void					instance(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const std::vector<SpatialTransform>&
										a_rTransformArray)					=0;

};

} // namespace

#endif /* __surface_SurfaceAccessibleI_h__ */
